/*
 * Copyright (c) 2008, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stdlib.h>
#include <stdarg.h>
#include <clutter/clutter.h>
#include <clutter-box2d.h>
#include <clutter-cairo.h>
#include <clutter/x11/clutter-x11.h>

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>

#include "moblin-film.h"
#include "moblin-category-box.h"
#include "moblin-declare.h"
#include "tween.h"

#include <genesis/genesis.h>

static gint full_screen = 1;
static gint always_on_top = 0;
static gint hide_cursor = 0;
static const ClutterColor black = { 0x00, 0x00, 0x00, 0xff };
static const ClutterColor white = { 0xff, 0xff, 0xff, 0xff };
static GenesisController *controller = NULL;
static ClutterActor *category_film = NULL;
static ClutterActor *active_category = NULL;
static ClutterActor *active_category_box = NULL;
typedef enum {
  VIEW_RANDOM = 1,
} VIEWMODE;
static VIEWMODE view_mode = VIEW_RANDOM;

typedef enum {
  LABEL_NO_FLAGS        = 0,
  LABEL_ACTIVE          = (1 << 0),
  LABEL_WITH_CIRCLE     = (1 << 1),
  LABEL_ALIGN_CENTER    = (1 << 2),
  LABEL_ALIGN_LEFT      = (1 << 3),
  LABEL_ALIGN_RIGHT     = (1 << 4),
  LABEL_ALIGN_MASK      = (LABEL_ALIGN_LEFT | 
			   LABEL_ALIGN_RIGHT | 
			   LABEL_ALIGN_CENTER),
  LABEL_FONT_SMALL      = (1 << 5),
  LABEL_FONT_BOLD       = (1 << 6),
} LABEL_DRAW_FLAGS;

guint moblin_label_get_width(const gchar *text, LABEL_DRAW_FLAGS flags)
{
  ClutterActor *actor = clutter_cairo_new(16, 16);
  cairo_t *cr = clutter_cairo_create(CLUTTER_CAIRO(actor));
  cairo_text_extents_t extents;

  cairo_select_font_face(cr, "Arial", 
			 CAIRO_FONT_SLANT_NORMAL,
			 flags & LABEL_FONT_BOLD ?
			 CAIRO_FONT_WEIGHT_BOLD :
			 CAIRO_FONT_WEIGHT_NORMAL);
  
  if (flags & LABEL_FONT_SMALL)
    cairo_set_font_size(cr, 10.0);
  else
    cairo_set_font_size(cr, 14.0);
  cairo_set_line_width(cr, 0.5);
  cairo_text_extents(cr, text, &extents);

  cairo_destroy(cr);
  clutter_actor_destroy(actor);

  return extents.width;
}

void moblin_draw_label(ClutterActor *actor, const gchar *text, 
		       LABEL_DRAW_FLAGS flags)
{
  cairo_t *cr;
  gint radius = 8;
  cairo_pattern_t *pat;
  cairo_text_extents_t extents;
  gdouble px;
  guint w, h;

  cr = clutter_cairo_create(CLUTTER_CAIRO(actor));

  cairo_set_operator(cr, CAIRO_OPERATOR_CLEAR);
  cairo_paint(cr);
  cairo_set_operator(cr, CAIRO_OPERATOR_OVER);

  clutter_actor_get_size(actor, &w, &h);

  px = h * 7.0 / 20.0;

  /* start the top */
  cairo_move_to(cr, radius + 0.5, 0.5);
	
  /* draw top */
  cairo_line_to(cr, w - radius + 0.5, 0.5);
  cairo_arc(cr, w - radius - 0.5, radius + 0.5, radius, 
	    3.0 * MOBLIN_PI / 2.0, 0.0);

  /* draw right */
  cairo_line_to(cr, w - 0.5, h - radius - 0.5);
  cairo_arc(cr, w - radius - 0.5, h - radius - 0.5, radius, 
	    0.0, MOBLIN_PI / 2.0);

  /* draw bottom */
  cairo_line_to(cr, radius + 0.5, h - 0.5);
  cairo_arc(cr, radius + 0.5, h - radius - 0.5, radius, 
	    MOBLIN_PI / 2.0, MOBLIN_PI);

  /* draw left */
  cairo_line_to(cr, + 0.5, 0.5 + radius);
  cairo_arc(cr, radius + 0.5, 0.5 + radius, radius,
	    MOBLIN_PI, 3.0 * MOBLIN_PI / 2.0);
  
  cairo_close_path (cr);

  /* Fill the path using a gradient if this item is active */
  if (flags & LABEL_ACTIVE) {
    pat = cairo_pattern_create_linear(0, 0, w, h);
    cairo_pattern_add_color_stop_rgba(pat, 0.1, 0.0, 0.0, 0.0, 0.4);
    cairo_pattern_add_color_stop_rgba(pat, 1.0, 0.7, 0.7, 0.7, 0.8);
    cairo_set_source(cr, pat);
    cairo_fill_preserve(cr);
    cairo_pattern_destroy(pat);
  } else {
    cairo_set_source_rgba(cr, 0, 0, 0, 0.40);
    cairo_fill_preserve(cr);
  }

  /* Draw the border around the app -- thickness varies if ACTIVE  */
  cairo_set_source_rgba(cr, 1, 1, 1, 0.75);
  cairo_set_line_width(cr, flags & LABEL_ACTIVE ? 3.0 : 2.0);
  cairo_stroke(cr);

  /* Draw a circle */
  if (flags & LABEL_WITH_CIRCLE) {
    cairo_arc(cr, w / 2, h / 2, px + 3, 0, 2.0 * MOBLIN_PI);
    cairo_set_source_rgba(cr, 1, 1, 1, 0.5);
    cairo_fill_preserve(cr);
    
    cairo_set_source_rgba(cr, 0, 0, 0, 1.0);
    cairo_set_line_width(cr, 3.0);
    cairo_stroke_preserve(cr);
  
    cairo_set_line_width(cr, 2.0);
    cairo_set_source_rgba(cr, 1, 1, 1, 1.0);
    cairo_stroke(cr);
  }

  /* Draw our label in the center of the actor */
  cairo_select_font_face(cr, "Arial", 
			 CAIRO_FONT_SLANT_NORMAL,
			 flags & LABEL_FONT_BOLD ?
			 CAIRO_FONT_WEIGHT_BOLD :
			 CAIRO_FONT_WEIGHT_NORMAL);
  
  if (flags & LABEL_FONT_SMALL)
    cairo_set_font_size(cr, 10.0);
  else
    cairo_set_font_size(cr, 14.0);
  cairo_set_line_width(cr, 0.5);
  cairo_text_extents(cr, text, &extents);

  if ((flags & LABEL_ALIGN_MASK) == LABEL_ALIGN_CENTER)
    cairo_move_to(cr, (w - extents.width) / 2 - 1, (h + extents.height) / 2);
  else
    cairo_move_to(cr, 8, (h + extents.height) / 2);
  cairo_text_path(cr, text);
  
  cairo_set_source_rgba(cr, 1, 1, 1, 1.0);
  cairo_fill_preserve(cr);
/*
  cairo_set_line_width(cr, 3.0);
  cairo_set_source_rgba(cr, 0, 0, 0, 1);
  cairo_stroke_preserve(cr);
  cairo_set_line_width(cr, 2.0);
  cairo_set_source_rgba(cr, 1, 1, 1, 1);
  cairo_stroke(cr);
*/
  cairo_destroy(cr);
}

ClutterActor *moblin_slot_new(guint width, guint height, 
			      const gchar *text, const gchar *icon_path,
			      LABEL_DRAW_FLAGS flags)
{
  guint w, h;
  cairo_t *cr;
  ClutterActor *app;
  ClutterActor *group;
  ClutterActor *icon = NULL;

  group = clutter_group_new();
  clutter_actor_set_opacity(group, 0xff);
  clutter_actor_set_position(group, 0, 10);

  app = clutter_cairo_new(width, height);
  clutter_container_add_actor(CLUTTER_CONTAINER(group), app);

  if (icon_path) {
    icon = clutter_texture_new_from_file(icon_path, NULL);
    clutter_actor_set_size(icon, 64, 64);
  } 

  if (!icon) {
    icon = clutter_cairo_new(64, 64);
    moblin_draw_label(icon, "?", LABEL_WITH_CIRCLE | LABEL_ALIGN_CENTER);
  }

  clutter_actor_get_size(app, &w, &h);
  clutter_actor_set_position(icon, w - 68, 4);
  clutter_container_add_actor(CLUTTER_CONTAINER(group), icon);
  
  moblin_draw_label(app, text, flags);

  return group;
}

char *find_icon_path(const char *path, const char *name)
{
    DIR *dir;
    struct dirent *entry;
    char *match = NULL;

    dir = opendir(path);
    if (!dir)
      goto skip_path;
    while (entry = readdir(dir)) {
      if (strcasestr(entry->d_name, name)) {
	struct stat st;
	asprintf(&match, "%s/%s", path, entry->d_name);
	stat(match, &st);
	if (S_ISDIR(st.st_mode)) {
	  free(match);
	  match = NULL;
	  continue;
	}

	break;
      }
    }
    closedir(dir);
skip_path:
    return match;
}


#define ICON_THEMES "Moblin:hicolor"
#define ICON_SIZES "scalable:128x128:48x48:32x32"
#define ICON_DIRS  "/usr/share"
#define MAX_SEARCH 12

const char *icon_themes[MAX_SEARCH] = {
  NULL
};

const char *icon_sizes[MAX_SEARCH] = {
  NULL
};

const char *icon_dirs[MAX_SEARCH] = {
  NULL
};

void fill_env_fields(const char *name, const char *value, const char *fields[])
{
  char *env = getenv(name);
  guint i;
  char *tok;

  if (env)
    env = g_strdup(env);
  else
    env = g_strdup(value);

  i = 0;
  tok = strtok(env, ":");
  while (tok && i < MAX_SEARCH) {
    fields[i++] = strdup(tok);
    tok = strtok(NULL, ":");
  }

  free(env);
}

void fill_search_fields() 
{
  fill_env_fields("ICON_THEMES", ICON_THEMES, icon_themes);
  fill_env_fields("ICON_SIZES", ICON_SIZES, icon_sizes);
  fill_env_fields("XDG_DATA_DIRS", ICON_DIRS, icon_dirs);
}

char *find_icon(char *type, const char *name)
{
  int i, j, k;
  char *match;
  char *base_name;
  char *path;

  /* First walk through all of the icon search directories to see if a
   * custom theme has provided an icon... */
  i = 0;
  while (icon_dirs[i]) {
    j = 0;
    while (icon_themes[j]) {
      k = 0;
      while (icon_sizes[k]) {
	asprintf(&path, "%s/icons/%s/%s/%s", 
		 icon_dirs[i], icon_themes[j], icon_sizes[k], type);
	match = find_icon_path(path, name);
	free(path);
	if (match)
	  return match;
	k++;
      }
      j++;
    }
    i++;
  }

  /* If the "name" is a path, search in the directory name part of the
   * path for a file name including the base name part of the path */
  base_name = strrchr(name, '/');
  if (base_name) {
    *base_name = '\0';
    match = find_icon_path(name, base_name + 1);
    *base_name = '/';
    if (match)
      return match;
  }
 

  /* Finally look in /usr/share/pixmaps for just the name */
  match = find_icon_path("/usr/share/pixmaps", name);
  if (match)
    return match;

  return NULL;
}

gchar *find_application_icon(gchar *name)
{
  gchar *icon;

  if (!access(name, R_OK))
    return g_strdup(name);
  
  icon = find_icon("apps", name);
  if (!icon)
    printf("No icon for application '%s'\n", name);
  printf("Using %s\n", icon);
  return icon;
}

gchar *find_category_icon(gchar *name)
{
  gchar *icon;

  if (!access(icon, R_OK))
    return g_strdup(icon);

  icon = find_icon("categories", name);
  if (!icon)
    printf("No icon for category '%s'\n", name);
  printf("Using %s\n", icon);
  return icon;
}

static glong moblin_app_movement(ClutterActor *actor, gint x, gint y)
{
  glong dx;
  glong dy;

  dx = x - (glong)g_object_get_data(G_OBJECT(actor), "moblin-app-x");
  dy = y - (glong)g_object_get_data(G_OBJECT(actor), "moblin-app-y");

  return (dx * dx + dy * dy);
}

static glong moblin_app_click_movement(ClutterActor *actor, gint x, gint y)
{
  glong dx;
  glong dy;

  dx = x - (glong)g_object_get_data(G_OBJECT(actor), "moblin-app-click-x");
  dy = y - (glong)g_object_get_data(G_OBJECT(actor), "moblin-app-click-y");

  return (dx * dx + dy * dy);
}

void moblin_actor_get_stage_relative_position(ClutterActor *actor, 
					      gint *x, gint *y)
{
  gint tmp_x, tmp_y;
  ClutterActor *parent;

  clutter_actor_get_position(actor, x, y);
  parent = clutter_actor_get_parent(actor);
  while (parent) {
    clutter_actor_get_position(parent, &tmp_x, &tmp_y);
    *x += tmp_x;
    *y += tmp_y;
    parent = clutter_actor_get_parent(parent);
  }
}

static gboolean moblin_app_hover(ClutterActor *actor)
{
  ClutterActor *label;

  gint x, y;
  if (!g_object_get_data(G_OBJECT(actor), "moblin-app-click-timeout")) {
    tween(actor, 100, "scale", 1.0, NULL);
    return FALSE;
  }
  g_object_set_data(G_OBJECT(actor), "moblin-app-click-timeout", NULL);
  clutter_actor_get_position(actor, &x, &y);
  if (moblin_app_movement(actor, x, y) > 25)
    return FALSE;

  label = g_object_get_data(G_OBJECT(actor), "moblin-app-label");
  if (!label) {
    guint w, h, text_width;
    gint x, y;
    char *text = g_object_get_data(G_OBJECT(actor), 
				   "moblin-app-entry-name");
    printf("Activating Label\n");

    text_width = moblin_label_get_width(text, LABEL_FONT_BOLD) + 16;
    label = clutter_cairo_new(text_width, 32);
    clutter_actor_get_size(actor, &w, &h);
    moblin_draw_label(label, text, LABEL_ALIGN_CENTER | LABEL_FONT_BOLD);
    g_object_set_data(G_OBJECT(actor), "moblin-app-label", label);
    clutter_actor_set_position(label, (w - text_width) / 2, -16);
    clutter_container_add_actor(CLUTTER_CONTAINER(actor), label);
  }

  clutter_actor_raise_top(actor);
  clutter_actor_show(label);

  return FALSE; /* Stop execution */
}


gboolean moblin_application_clicked(ClutterActor *actor,
				    ClutterButtonEvent *event,
				    gpointer data)
{
  gint x, y;
  if (event->click_count != 1)
    return TRUE;
  clutter_actor_get_position(actor, &x, &y);
  g_object_set_data(
    G_OBJECT(actor), "moblin-app-click-timeout", 
    (void*)(glong)g_timeout_add(500, (GSourceFunc)moblin_app_hover, actor));
  g_object_set_data(G_OBJECT(actor), "moblin-app-x", 
		    (void*)(glong)x);
  g_object_set_data(G_OBJECT(actor), "moblin-app-y", 
		    (void*)(glong)y);
  g_object_set_data(G_OBJECT(actor), "moblin-app-click-x", 
		    (void*)(glong)event->x);
  g_object_set_data(G_OBJECT(actor), "moblin-app-click-y", 
		    (void*)(glong)event->y);
  g_object_set_data(G_OBJECT(actor), "moblin-app-click-time", 
		    (void*)(gulong)event->time);

  tween(actor, 100, "scale", 0.95, NULL);

  return TRUE;
}

gboolean moblin_application_released(ClutterActor *actor,
				     ClutterButtonEvent *event,
				     gpointer data)
{
  static guint last_time = 0;
  ClutterActor *label;
  gulong elapsed;
  
  if (event->time == last_time)
    return FALSE;

  last_time = event->time;

  g_object_set_data(G_OBJECT(actor), "moblin-app-click-timeout", NULL);
  tween(actor, 100, "scale", 1.0, NULL);
  label = g_object_get_data(G_OBJECT(actor), "moblin-app-label");
  if (label)
    clutter_actor_hide(label);

  elapsed = event->time;
  elapsed -= (gulong)g_object_get_data(G_OBJECT(actor), 
				       "moblin-app-click-time");
  if (moblin_app_click_movement(actor, event->x, event->y) < 10 &&
      elapsed < 250) {
    gchar *name;
    gchar *old_category = NULL;

    name = g_object_get_data(G_OBJECT(actor), "moblin-app-entry-name");
    if (!name)
      return FALSE;
    genesis_controller_start_app_from_name(controller, name);

    return TRUE;
  }

  return FALSE;
}

ClutterActor *moblin_app_new(ClutterActor *app_box,
			     guint width, guint height, 
			     GenesisAppEntry *entry)
{
  ClutterActor *icon = NULL;
  gchar *icon_path = find_application_icon(genesis_app_entry_get_icon(entry));
  GList *list;
  gint x, y;
  guint w, h;
  ClutterActor *last_icon;
  ClutterActor *group;

  group = clutter_group_new();

  if (icon_path) {
    icon = clutter_texture_new_from_file(icon_path, NULL);
    if (icon)
      clutter_actor_set_size(icon, width, height);
    g_free(icon_path);
  } 

  if (!icon) {
    icon = clutter_cairo_new(width, height);
    moblin_draw_label(icon, genesis_app_entry_get_name(entry), 
		      LABEL_WITH_CIRCLE | LABEL_ALIGN_CENTER | 
		      LABEL_FONT_SMALL);
  }

  clutter_container_add_actor(CLUTTER_CONTAINER(group), icon);

  list = clutter_container_get_children(CLUTTER_CONTAINER(app_box));
  list = g_list_last(list);

  x = 4; 
  y = 4;
  w = 0; 
  h = 0;
  
  if (list) {
    last_icon = CLUTTER_ACTOR(list->data);
    if (g_object_get_data(G_OBJECT(last_icon), "moblin-app-entry-name")) {
      clutter_actor_get_size(last_icon, &w, &h);
      clutter_actor_get_position(last_icon, &x, &y);
    }
  }

  x += w + 4;
  if (x + width > clutter_actor_get_width(app_box) - width) {
    x = 4;
    y += h + 8;
  }

  if (view_mode == VIEW_RANDOM)
    clutter_actor_set_position(
      group,
      rand() % (clutter_actor_get_width(app_box) - width),
      rand() % (clutter_actor_get_height(app_box) - height));
  else
    clutter_actor_set_position(group, x, y);


  clutter_container_add_actor(CLUTTER_CONTAINER(app_box), group);
  clutter_container_child_set (CLUTTER_CONTAINER(app_box), group,
			       "manipulatable", TRUE,
			       "mode",          CLUTTER_BOX2D_DYNAMIC,
			       NULL);
  clutter_box2d_set_simulating(CLUTTER_BOX2D(app_box), TRUE);

  g_object_set_data(G_OBJECT(group), "moblin-app-entry-name",
		    g_strdup(genesis_app_entry_get_name(entry)));
  g_signal_connect(G_OBJECT(group), "button-press-event",
		   G_CALLBACK(moblin_application_clicked), NULL);
  g_signal_connect(G_OBJECT(group), "button-release-event",
		   G_CALLBACK(moblin_application_released), NULL);
  
  return icon;
}

static gboolean moblin_destroy_old_category(gpointer data)
{
  clutter_actor_destroy(CLUTTER_ACTOR(data));
  return FALSE; /* Stop execution */
}

static void moblin_category_open(gchar *name)
{
  ClutterActor *app_box = g_object_new(MOBLIN_TYPE_CATEGORY_BOX, NULL);
  GList *apps;

  g_object_set_data(G_OBJECT(app_box), 
		    "moblin-category-name", g_strdup(name));
  
  guint left;
  left = clutter_actor_get_width(category_film) +
    clutter_actor_get_x(category_film) + 16;
  clutter_actor_set_size(app_box, CLUTTER_STAGE_WIDTH() - left - 16,
			 CLUTTER_STAGE_HEIGHT() - 32);
  clutter_actor_set_position(app_box, left + 
			     (CLUTTER_STAGE_WIDTH() - left) / 2, 
			     CLUTTER_STAGE_HEIGHT() / 2);
  clutter_box2d_set_simulating(CLUTTER_BOX2D(app_box), TRUE);
  clutter_container_add_actor(CLUTTER_CONTAINER(
				clutter_stage_get_default()), app_box);

  apps = genesis_controller_get_applications_by_category(controller, name);
  if (apps)
    apps = g_list_sort(apps, (GCompareFunc)strcasecmp);
  while (apps) {
    GenesisAppEntry *entry = GENESIS_APP_ENTRY(apps->data);
    ClutterActor *app;
    app = moblin_app_new(app_box, 64, 64, entry);
    apps = apps->next;
  }
  if (apps)
    g_list_free(apps);

  clutter_actor_set_opacity(app_box, 0);
  clutter_actor_set_scale(app_box, 0.25, 0.25);
  tween(app_box, 500, 
	"scale", 1.0, 
	"opacity", 1.0 * 0xff, 
	"x", 1.0 * left,
	"y", 16.0,
	NULL);

  /* hack -- todo put this into storage and manage it correctly*/
  active_category_box = app_box;
}

gboolean moblin_category_clicked(ClutterActor *actor,
				 ClutterButtonEvent *event,
				 gpointer data)
{
  if (event->click_count == 1) {
    g_object_set_data(G_OBJECT(actor), "moblin-app-click-x", 
		      (void*)(glong)event->x);
    g_object_set_data(G_OBJECT(actor), "moblin-app-click-y", 
		      (void*)(glong)event->y);
    g_object_set_data(G_OBJECT(actor), "moblin-app-click-time", 
		      (void*)(gulong)event->time);
  }
  return FALSE;
}

gboolean moblin_category_released(ClutterActor *actor,
				  ClutterButtonEvent *event,
				  gpointer data)
{
  gulong elapsed;

  elapsed = event->time;
  elapsed -= (gulong)g_object_get_data(G_OBJECT(actor), 
				       "moblin-app-click-time");

  if (moblin_app_click_movement(actor, event->x, event->y) < 10 &&
      elapsed < 500) {
    gchar *category;
    gchar *old_category = NULL;
    gint x, y;

    category = g_object_get_data(G_OBJECT(actor), "moblin-category-name");
    if (!category)
      return FALSE;

    if (active_category_box) {
      old_category = g_object_get_data(G_OBJECT(active_category_box), 
				       "moblin-category-name");
	if (!strcmp(old_category, category))
	  return FALSE;
    }
    
    printf("Open %s\n", category);
    clutter_actor_get_position(active_category_box, &x, &y);
    tween(active_category_box, 500, 
	  "scale", 1.5,
	  "opacity", 0.0 * 0xff, 
	  "x", (1.0 * x) - 0.25 * clutter_actor_get_width(active_category_box),
	  "y", (1.0 * y) - 0.25 * clutter_actor_get_height(active_category_box),
	  NULL);
    g_timeout_add(700, moblin_destroy_old_category, active_category_box);
    if (active_category_box) {
      GList *children = clutter_container_get_children(
	CLUTTER_CONTAINER(active_category));
      moblin_draw_label(children->data, old_category, LABEL_NO_FLAGS);
      active_category_box = NULL;
    }
    
    active_category = actor;
    {
      GList *children = clutter_container_get_children(
	CLUTTER_CONTAINER(actor));
      moblin_draw_label(children->data, category, LABEL_ACTIVE);
    }

    moblin_category_open(category);
  }
  return TRUE;
}

static void moblin_key_release(ClutterStage           *stage,
			       ClutterKeyEvent        *kev,
			       gpointer                user_data)
{
  switch (clutter_key_event_symbol (kev)) {
      case CLUTTER_Escape:
        clutter_main_quit ();
        break;
    }
}

int main(int argc, char **argv)
{
  ClutterActor *stage, *background;
  ClutterActor *app;
  gint i;
  GList *categories;
  GList *first;

  fill_search_fields();

  /* gtk init needed by libwnck used by genesis */
  gtk_init(&argc, &argv);
  clutter_init(&argc, &argv);
  
  controller = genesis_controller_get_singleton();
  if (!controller) {
    g_warning ("Failed to get genesis controller singleton.\n");
    return -1;
  }

  stage = clutter_stage_get_default();
  clutter_actor_set_size(stage, SCREEN_WIDTH, SCREEN_HEIGHT);
  if (full_screen)
    clutter_stage_fullscreen(CLUTTER_STAGE(stage));
/*  if (always_on_top)
    clutter_stage_above(CLUTTER_STAGE(stage));*/
  if (hide_cursor)
    clutter_stage_hide_cursor(CLUTTER_STAGE(stage));
  
  background = clutter_texture_new_from_file("background.jpg", NULL);
  if (background) {
    clutter_actor_set_size(background, CLUTTER_STAGE_WIDTH(),
			   CLUTTER_STAGE_HEIGHT());
    clutter_actor_set_position(background, 0, 0);
    clutter_container_add_actor(CLUTTER_CONTAINER(stage), 
				background);
  } else {
    clutter_stage_set_color(CLUTTER_STAGE(stage), &black);
  }
  
  category_film = g_object_new(MOBLIN_TYPE_FILM, NULL);
  clutter_actor_set_size(category_film, 200, CLUTTER_STAGE_HEIGHT());
  clutter_actor_set_position(category_film, 32, 0);

  categories = genesis_controller_get_categories(controller);
  if (categories)
    categories = g_list_sort(categories, (GCompareFunc)strcasecmp);
  first = categories;
  while (categories) {
    GList *apps;
    apps = genesis_controller_get_applications_by_category(controller, 
							   categories->data);
    if (!apps) {
      printf("Skipping %s (no apps)\n", categories->data);
      categories = categories->next;
      continue;
    }

    app = moblin_slot_new(200, 32, categories->data, 
			  find_category_icon(categories->data),
			  (first == categories ? 
			   LABEL_ACTIVE : LABEL_NO_FLAGS));

    if (first == categories)
      active_category = app;

    clutter_container_add_actor(CLUTTER_CONTAINER(category_film), app);
    g_object_set_data(G_OBJECT(app), "moblin-category-name", 
		      g_strdup(categories->data));

    g_signal_connect(G_OBJECT(app), "button-press-event",
		     G_CALLBACK(moblin_category_clicked), NULL);
    g_signal_connect(G_OBJECT(app), "button-release-event",
		     G_CALLBACK(moblin_category_released), NULL);
    
    categories = categories->next;
  }

  if (first)
    moblin_category_open(first->data);
  
  g_list_free(categories);

/*
  g_signal_connect (G_OBJECT(controller), "app-entry-updated",
  G_CALLBACK(moblin_app_entry_updated), NULL);
*/  
  clutter_box2d_set_simulating(CLUTTER_BOX2D(category_film), TRUE);
  
  clutter_container_add_actor(CLUTTER_CONTAINER(stage), category_film);

  clutter_actor_show_all(stage);

  g_signal_connect (stage,
		    "key-release-event",
		    G_CALLBACK(moblin_key_release),
		    NULL);

  clutter_main();
  
  return 0;
}
