/*
 * Copyright (c) 2008, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <clutter/clutter.h>
#include <cogl/cogl.h>
#include "clutter-box2d.h"
#include <string.h>
#include <math.h>
#include "moblin-category-box.h"
#include "moblin-declare.h"

G_DEFINE_TYPE(MoblinCategory_Box, moblin_category_box, CLUTTER_TYPE_BOX2D);

#define MOBLIN_CATEGORY_BOX_GET_PRIVATE(obj) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MOBLIN_TYPE_CATEGORY_BOX, MoblinCategory_BoxPrivate))

enum {
  BOUNDARY_LEFT = 0,
  BOUNDARY_TOP,
  BOUNDARY_RIGHT,
  BOUNDARY_BOTTOM
};

struct _MoblinCategory_BoxPrivate
{
  ClutterActor *boundaries[4];
  guint w;
  guint h;
};

typedef struct MoblinCategory_BoxAppData
{
} MoblinCategory_BoxAppData;

static void moblin_category_box_resize(ClutterActor *actor,
				       GParamSpec *param_spec,
				       gpointer data)
{
  MoblinCategory_BoxPrivate *priv;
  gint gap;
  guint h, w;
  ClutterActor *bound;
  ClutterActor *bounds[4];
  gint i;
  ClutterContainer *container = CLUTTER_CONTAINER(actor);
  MoblinCategory_BoxAppData *box_data;

  priv = MOBLIN_CATEGORY_BOX_GET_PRIVATE(actor);
  clutter_actor_get_size(CLUTTER_ACTOR(container), &w, &h);

  if (w == priv->w && h == priv->h)
    return;

  priv->w = w;
  priv->h = h;

  printf("resize..\n");
  /* Create top, left, bottom, and right boundary boxes */
  for (i = 0; i < 4; i++) {
    bounds[i] = clutter_rectangle_new();
    clutter_actor_set_opacity(bounds[i], 0xff * 0.00);
  }
  
  /* Size and position the four boundaries */
  bound = bounds[BOUNDARY_LEFT];
  clutter_actor_set_size(bound, 10, h);
  clutter_actor_set_position(bound, -10, 0);
    
  bound = bounds[BOUNDARY_TOP];
  clutter_actor_set_size(bound, 10 * 2 + w, 10);
  clutter_actor_set_position(bound, -10, -10);
    
  bound = bounds[BOUNDARY_RIGHT];
  clutter_actor_set_size(bound, 10, h);
  clutter_actor_set_position(bound, w, 0);
    
  bound = bounds[BOUNDARY_BOTTOM];
  clutter_actor_set_size(bound, 10 * 2 + w, 10);
  clutter_actor_set_position(bound, -10, h);

  /* Add the new boundaries to Box2D */
  for (i = 0; i < 4; i++) {
    clutter_container_add_actor(container, bounds[i]);
    clutter_container_child_set(container, bounds[i], 
				"mode", CLUTTER_BOX2D_STATIC, NULL);
  }

  /* Remove the older boundaries if they exist and cache the new ones */
  for (i = 0; i < 4; i++) {
    if (priv->boundaries[i]) {
      clutter_actor_hide(priv->boundaries[i]);
      clutter_container_remove_actor(container, priv->boundaries[i]);
      
    }
    priv->boundaries[i] = bounds[i];
  }
}

static GObject *moblin_category_box_constructor(
  GType                  type,
  guint                  n_params,
  GObjectConstructParam *params)
{
  GObject      *object;
  ClutterActor *box2d;
  MoblinCategory_Box *category_box;
  MoblinCategory_BoxPrivate *priv;
  ClutterVertex gravity = { 0, 0 };

  object = G_OBJECT_CLASS(moblin_category_box_parent_class)->constructor(
    type, n_params, params);
  box2d  = CLUTTER_ACTOR(object);
  category_box = MOBLIN_CATEGORY_BOX(object);
  priv   = category_box->priv;

  g_object_set (box2d, "gravity", &gravity, NULL);

  g_signal_connect_after(object, "notify::allocation",
			 G_CALLBACK (moblin_category_box_resize), NULL);

  return object;
}

static void moblin_category_box_class_init (MoblinCategory_BoxClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->constructor  = moblin_category_box_constructor;
  g_type_class_add_private (gobject_class, sizeof (MoblinCategory_BoxPrivate));

}

static void moblin_category_box_init (MoblinCategory_Box *self)
{
  MoblinCategory_BoxPrivate *priv;
  self->priv = priv = MOBLIN_CATEGORY_BOX_GET_PRIVATE (self);
}
