/*
 * Copyright (c) 2008, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implement a film strip UI pattern
 *
 * Author: James Ketrenos <jketreno@linux.intel.com>
 *
 */
#include <clutter/clutter.h>
#include <cogl/cogl.h>
#include "clutter-box2d.h"
#include <string.h>
#include <math.h>
#include "moblin-film.h"
#include "moblin-declare.h"

#define CELL_PADDING 8
#define CELL_HEIGHT  80
#define CELL_WIDTH   120

#define FREQUENCY 0.0
#define DAMPING 0.0

G_DEFINE_TYPE(MoblinFilm, moblin_film, CLUTTER_TYPE_BOX2D);

#define MOBLIN_FILM_GET_PRIVATE(obj) \
   (G_TYPE_INSTANCE_GET_PRIVATE ((obj), MOBLIN_TYPE_FILM, MoblinFilmPrivate))

enum {
  BOUNDARY_LEFT = 0,
  BOUNDARY_TOP,
  BOUNDARY_RIGHT,
  BOUNDARY_BOTTOM
};

struct _MoblinFilmPrivate
{
  guint max_width;
  guint next_id;
  guint cell_count;
  guint strip_height;
  GList *cells;
  ClutterActor *boundaries[4]; /* l, t, r, b */
  ClutterActor *segments[3];
  ClutterBox2DJoint *segment_joints[2];
};

typedef struct MoblinFilmCellData
{
  gint id;
  ClutterActor *self;
  ClutterActor *prev_cell;
  ClutterActor *next_cell;
  ClutterBox2DJoint *distance_joint[2];
  gint notify_handler;
} MoblinFilmCellData;

static void moblin_film_join_cells(ClutterActor *actor1,
				   ClutterActor *actor2)
{
  ClutterVertex v[4];
  ClutterActor *parent = clutter_actor_get_parent(actor1);
  MoblinFilmCellData *cell1;
  ClutterBox2D *box2d = CLUTTER_BOX2D(parent);

  cell1 = g_object_get_data(G_OBJECT(actor1), "moblin-film-cell-data");

  if (cell1->distance_joint[0]) {
    clutter_box2d_joint_destroy(cell1->distance_joint[0]);
    clutter_box2d_joint_destroy(cell1->distance_joint[1]);
  }

  v[0].x = CLUTTER_UNITS_FROM_FLOAT(0.5 * clutter_actor_get_width(actor1));
  v[0].y = CLUTTER_UNITS_FROM_INT(clutter_actor_get_height(actor1));

  v[1].x = v[0].x;
  v[1].y = 0.0;

  v[2].x = v[0].x;
  v[2].y = CLUTTER_UNITS_FROM_FLOAT(0.5 * clutter_actor_get_height(actor1));

  v[3].x = CLUTTER_UNITS_FROM_FLOAT(0.5 * clutter_actor_get_width(actor2));
  v[3].y = CLUTTER_UNITS_FROM_FLOAT(0.5 * clutter_actor_get_height(actor2));

  cell1->distance_joint[0] = clutter_box2d_add_distance_joint(
    box2d, actor1, actor2, &v[0], &v[1], CELL_PADDING, FREQUENCY, DAMPING);
  cell1->distance_joint[1] = clutter_box2d_add_distance_joint(
    box2d, actor1, actor2, &v[2], &v[3], 
    clutter_actor_get_height(actor1) / 2.0 + 
    CELL_PADDING +
    clutter_actor_get_height(actor2) / 2.0, FREQUENCY, DAMPING);
}

static void moblin_film_detach(MoblinFilm *film,
			       ClutterActor *actor)
{
  gint x, y;
  MoblinFilmCellData *cell_data, *tmp_cell_data;
  
  cell_data = g_object_get_data(G_OBJECT(actor), 
				"moblin-film-cell-data");
  if (!cell_data)
    return;
  
  if (cell_data->prev_cell) {
    tmp_cell_data = g_object_get_data(G_OBJECT(cell_data->prev_cell),
				      "moblin-film-cell-data");
    tmp_cell_data->next_cell = cell_data->next_cell;
    if (tmp_cell_data->distance_joint[0]) {
      clutter_box2d_joint_destroy(tmp_cell_data->distance_joint[0]);
      clutter_box2d_joint_destroy(tmp_cell_data->distance_joint[1]);
      tmp_cell_data->distance_joint[0] = NULL;
      tmp_cell_data->distance_joint[1] = NULL;
    }
  }

  if (cell_data->next_cell) {
    tmp_cell_data = g_object_get_data(G_OBJECT(cell_data->next_cell),
				      "moblin-film-cell-data");
    tmp_cell_data->prev_cell = cell_data->prev_cell; 
    cell_data->distance_joint[0] = NULL;
    cell_data->distance_joint[1] = NULL;
  }
  
  /* Pin the previous cell to the next cell (if they both exist) */
  if (cell_data->prev_cell && cell_data->next_cell)
    moblin_film_join_cells(cell_data->prev_cell, cell_data->next_cell);

  cell_data->next_cell = NULL;
  cell_data->prev_cell = NULL;
}

/*
 * Attach this actor to the bottom of the previous cell
 */
static void moblin_film_attach_to_prev(MoblinFilm *film,
				       ClutterActor *actor)
{
  gint x, y;
  MoblinFilmCellData *cell;
  MoblinFilmPrivate *priv;

  priv = MOBLIN_FILM_GET_PRIVATE(film);
  
  cell = g_object_get_data(G_OBJECT(actor), "moblin-film-cell-data");
  if (!cell || !cell->prev_cell)
    return;
  
  /* Move cell to be below the previous cell */
  clutter_actor_get_position(cell->prev_cell, &x, &y);
  y += clutter_actor_get_height(cell->prev_cell) + CELL_PADDING;
  clutter_actor_set_position(actor, 0, y);

  /* Add the joints */
  moblin_film_join_cells(cell->prev_cell, actor);
}


static void moblin_film_cell_notify_position(ClutterActor *actor,
					     GParamSpec   *pspec,
					     gpointer data)
{
  MoblinFilm *film;
  MoblinFilmPrivate *priv;
  gint h;
  gint x, y, y_tmp;
  gint cx, cy;
  ClutterActor *box2d;
  MoblinFilmCellData *cell;
  gint redraw = 0;

  film = MOBLIN_FILM(clutter_actor_get_parent(actor));
  priv = MOBLIN_FILM_GET_PRIVATE(film);
  box2d  = CLUTTER_ACTOR(film);
  cell = g_object_get_data(G_OBJECT(actor), "moblin-film-cell-data");

  h = clutter_actor_get_height(actor);
  clutter_actor_get_position(actor, &x, &y);

  y_tmp = y;


  /* If the actor somehow sprung to beyond the midpoint of its neighbor
   * cells then force it back into its slot... */
  if (cell->prev_cell) {
    gint cx, cy, ch;
    ch = clutter_actor_get_height(cell->prev_cell);
    clutter_actor_get_position(cell->prev_cell, &cx, &cy);
    if (y <= cy + ch / 2)
      y = (cy + ch / 2) + 1;
  }
  if (cell->next_cell) {
    gint cx, cy, ch;
    ch = clutter_actor_get_height(cell->next_cell);
    clutter_actor_get_position(cell->next_cell, &cx, &cy);
    if (y + h >= cy + ch / 2)
      y = (cy + ch / 2) - (h + 1);
  }

  if (y != y_tmp) {
    gint cx, cy, ch;

    /* Make sure y is not now outside of our bounding boxes... */
    clutter_actor_get_position(priv->boundaries[BOUNDARY_TOP], &cx, &cy);
    ch = clutter_actor_get_height(priv->boundaries[BOUNDARY_TOP]);
    if (y < cy + ch) 
      y = cy + ch;

    clutter_actor_get_position(priv->boundaries[BOUNDARY_BOTTOM], &cx, &cy);
    if (y + h > cy) 
      y = cy - h;
    
    clutter_actor_set_position(actor, x, y);

    redraw = 1;
  }

  if (clutter_actor_is_rotated(actor)) {
    clutter_actor_set_rotation(actor, CLUTTER_Z_AXIS, 0.0, 0, 0, 0);
    redraw = 1;
  }
  
  if (redraw)
    clutter_actor_queue_redraw(actor);
}

static void moblin_film_resize(MoblinFilm *film)
{
  MoblinFilmPrivate *priv;
  gint gap;
  guint h, w;
  ClutterActor *bound;
  ClutterActor *bounds[4];
  gint i;
  ClutterContainer *container = CLUTTER_CONTAINER(film);

  priv = MOBLIN_FILM_GET_PRIVATE(film);
  w = priv->max_width;
  h = CLUTTER_STAGE_HEIGHT();
  
  gap = priv->strip_height - h;
  if (gap < 0)
    gap = 0;
  
  /* Create top, left, bottom, and right boundary boxes */
  for (i = 0; i < 4; i++) {
    bounds[i] = clutter_rectangle_new();
    clutter_actor_set_opacity(bounds[i], 0xff * 0.00);
    g_object_set_data(G_OBJECT(bounds[i]), "moblin-film-boundary", (gpointer)1);
  }
  
  /* Size and position the four boundaries */
  bound = bounds[BOUNDARY_LEFT];
  clutter_actor_set_size(bound, 10, h + gap * 2);
  clutter_actor_set_position(bound, -10, -gap);
    
  bound = bounds[BOUNDARY_TOP];
  clutter_actor_set_size(bound, 10 * 2 + w, 10);
  clutter_actor_set_position(bound, -10, -gap - 10);
    
  bound = bounds[BOUNDARY_RIGHT];
  clutter_actor_set_size(bound, 10, h + gap * 2);
  clutter_actor_set_position(bound, w, -gap);
    
  bound = bounds[BOUNDARY_BOTTOM];
  clutter_actor_set_size(bound, 10 * 2 + w, 10);
  clutter_actor_set_position(bound, -10, h + gap);

  /* Add the new boundaries to Box2D */
  for (i = 0; i < 4; i++) {
    clutter_container_add_actor(container, bounds[i]);
    clutter_container_child_set(container, bounds[i], 
				"mode", CLUTTER_BOX2D_STATIC, NULL);
  }

  /* Remove the older boundaries if they exist and cache the new ones */
  for (i = 0; i < 4; i++) {
    if (priv->boundaries[i]) {
      clutter_actor_hide(priv->boundaries[i]);
      clutter_container_remove_actor(container, priv->boundaries[i]);
      
    }
    priv->boundaries[i] = bounds[i];
  }
}


static void moblin_film_actor_added(MoblinFilm   *film,
				    ClutterActor *actor,
				    gpointer      user_data)
{
  MoblinFilmPrivate *priv;
  MoblinFilmCellData *cell;
  ClutterActor *last;
  MoblinFilmCellData *last_cell;
  guint w, h;

  if (g_object_get_data(G_OBJECT(actor), "moblin-film-boundary"))
    return;

  cell = g_object_get_data(G_OBJECT(actor), 
			   "moblin-film-cell-data");
  if (!cell) {
    cell = g_new0(MoblinFilmCellData, 1);
    g_object_set_data(G_OBJECT(actor), 
		      "moblin-film-cell-data", cell);
    cell->self = actor;
  }
  
  priv = MOBLIN_FILM_GET_PRIVATE(film);

  /* Turn on Box2D for the actor before we hook it to a joint otherwise
   * it won't have the actor may not be initialized within Box2D */
  clutter_container_child_set(CLUTTER_CONTAINER(film), actor,
			      "manipulatable", TRUE,
			      "mode",          CLUTTER_BOX2D_DYNAMIC,
			      NULL);

  /* Find the last cell in the film strip */
  if (g_list_last(priv->cells)) {
    last = CLUTTER_ACTOR(g_list_last(priv->cells)->data);
    last_cell = g_object_get_data(G_OBJECT(last), 
				  "moblin-film-cell-data");
  } else {
    last = NULL;
    last_cell = NULL;
  }

  /* Add this cell to the ordered list of cells */
  g_object_ref(actor);
  priv->cells = g_list_append(priv->cells, actor);
  priv->cell_count++;

  clutter_actor_get_size(actor, &w, &h);
  priv->strip_height += h + CELL_PADDING;
  if (w > priv->max_width)
    priv->max_width = w;

  /* Hook this cell to the end of the film strip */
  cell->prev_cell = last;
  if (last_cell)
    last_cell->next_cell = actor;

  moblin_film_resize(film);

  /* Attach the joint */
  moblin_film_attach_to_prev(film, actor);

  cell->id = priv->next_id++;

  /* Tracking notify::y didn't always work (sometimes the notifications
   * would fall silent) so we track allocation instead */
  cell->notify_handler = g_signal_connect(
    G_OBJECT(actor), "notify::allocation",
    G_CALLBACK(moblin_film_cell_notify_position), actor);
}	


static void moblin_film_actor_removed (MoblinFilm       *film,
				       ClutterActor *actor,
				       gpointer      user_data)
{
  MoblinFilmPrivate *priv = MOBLIN_FILM_GET_PRIVATE(film);
  MoblinFilmCellData *cell;

  cell = g_object_get_data(G_OBJECT(actor), 
			   "moblin-film-cell-data");
  if (cell) {
    priv->cells = g_list_remove(priv->cells, actor);
    g_object_unref(actor);
    priv->strip_height -= clutter_actor_get_height(actor) + CELL_PADDING;
    moblin_film_detach(film, actor);
    if (cell->notify_handler) {
      g_signal_handler_disconnect(actor, cell->notify_handler);
      cell->notify_handler = 0;
    }
    priv->cell_count--;
  }
}

static GObject *moblin_film_constructor (GType                  type,
					 guint                  n_params,
					 GObjectConstructParam *params)
{
  GObject      *object;
  ClutterActor *box2d;
  MoblinFilm *film;
  MoblinFilmPrivate *priv;
  ClutterVertex gravity = { 0, 0 };

  object = G_OBJECT_CLASS(moblin_film_parent_class)->constructor(
    type, n_params, params);
  box2d  = CLUTTER_ACTOR(object);
  film = MOBLIN_FILM(object);
  priv   = film->priv;

  g_object_set (box2d, "gravity", &gravity, NULL);

  moblin_film_resize(film);

  g_signal_connect_after(object, "actor-added",
			 G_CALLBACK (moblin_film_actor_added), NULL);
  g_signal_connect(object, "actor-removed",
		   G_CALLBACK (moblin_film_actor_removed), NULL);

  return object;
}

static void moblin_film_class_init (MoblinFilmClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  gobject_class->constructor  = moblin_film_constructor;
  g_type_class_add_private (gobject_class, sizeof (MoblinFilmPrivate));

}

static void moblin_film_init (MoblinFilm *self)
{
  MoblinFilmPrivate *priv;
  self->priv = priv = MOBLIN_FILM_GET_PRIVATE (self);
}
