/*
 * Copyright (c) 2008, Intel Corporation
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Intel Corporation nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implement a film strip UI pattern
 *
 *
 *   ----
 *==========
 *  |    | <---- Cell
 *  | v  |
 *   -|--  \
 *    |   -->--- Cell Padding
 *   -|--  / 
 *  | ^  |
 *  |    | <---- Cell
 *  | v  |
 *   -|--  \
 *    |   -->--- Cell Padding
 *   -|--  / 
 *  | ^  |
 *==========
 *  |    | <---- Cell
 *   ----
 *
 * The film strip has N visible on screen frames and M 
 * potential actors to map to those frames.
 *
 * Box2D is used to provide the physics and manipulation of the
 * film strip.  
 *
 *        --------- -.
 *       |         |   \
 *       |         |    }---- Gap = max(0, film_roll_length - viewport_height)
 *       |         |   /
 *       |         | -'
 *       |  [ o ]  | -----------------------------------.
 *       |         | <------- Boundary Actors (STATIC)   \
 *       |  [ o ]  |                                       }- Film Roll Length
 *  ---------------------                                 /
 * |     |         |     | <- Viewport                    |
 * |     |  [ o ]  |     |                                |
 * |     |         |     |                                |
 * |     |  [ o ] <---------- Cell on Film Roll (DYNAMIC) |
 * |     |         |     |                                |
 *  ---------------------                                 /
 *       |  [ o ]  | ------------------------------------'
 *       |         | -.
 *       |         |   \
 *       |         |    }---- Gap = max(0, film_roll_length - viewport_height)
 *       |         |   /
 *        ---------  -'     
 *
 * As actors are added to the film strip, the boundary actors are resized
 * and positioned to provide the rail guides for the cells to slide within.
 *
 * Currently the film strip is limited in how many cells can be added due
 * to restrictions in Box2D.  Eventually the limit will be improved by
 * removing off-screen actors from the Box2D world and replacing them with a 
 * dynamically growing / shrinking place holder Actor for all off-screen
 * Actors above and below the viewport.
 */
#ifndef __MOBLIN_FILM_H__
#define __MOBLIN_FILM_H__

#include <glib-object.h>
#include <clutter/clutter-actor.h>
#include "clutter-box2d.h"

G_BEGIN_DECLS

#define MOBLIN_TYPE_FILM (moblin_film_get_type ())
#define MOBLIN_FILM(obj) (G_TYPE_CHECK_INSTANCE_CAST \
			  ((obj), MOBLIN_TYPE_FILM, MoblinFilm))

typedef struct _MoblinFilm         MoblinFilm;
typedef struct _MoblinFilmClass    MoblinFilmClass;
typedef struct _MoblinFilmPrivate  MoblinFilmPrivate;

struct _MoblinFilm
{
	ClutterBox2D       parent;
	MoblinFilmPrivate *priv;
};

struct _MoblinFilmClass 
{
	/*< private >*/
	ClutterBox2DClass parent_class;
}; 

GType  moblin_film_get_type (void) G_GNUC_CONST;

G_END_DECLS

#endif /* __MOBLIN_FILM_H__ */
